#include <fggl/util/service.h>
#include <spdlog/spdlog.h>

#include <fggl/data/model.hpp>
#include <fggl/gfx/camera.hpp>
#include <fggl/gfx/ogl/renderer.hpp>
#include <fggl/gfx/paint.hpp>
#include <fggl/math/triangulation.hpp>
#include <glm/ext/matrix_clip_space.hpp>
#include <glm/ext/matrix_transform.hpp>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <memory>

extern "C" {

/**
 * Convert an OpenGL source enum to a string.
 *
 * list of sources taken from table 20.1, GL Spec 4.5
 * @param source
 * @return string representing the source
 */
constexpr auto fggl_ogl_source(GLenum source) -> const char * {
	switch (source) {
	case GL_DEBUG_SOURCE_API: return "GL";
	case GL_DEBUG_SOURCE_SHADER_COMPILER: return "GLSL compiler";
	case GL_DEBUG_SOURCE_WINDOW_SYSTEM: return "Windowing System";
	case GL_DEBUG_SOURCE_THIRD_PARTY: return "Third Party";
	case GL_DEBUG_SOURCE_APPLICATION: return "Application";
	default:
	case GL_DEBUG_SOURCE_OTHER: return "Other";
	}
	assert(false);
	return "unknown";
}

/**
 * Convert an OpenGL type enum to a string.
 *
 * list of sources taken from table 20.2, GL Spec 4.5
 * @param type
 * @return string representing the type
 */
constexpr auto static fggl_ogl_type(GLenum type) -> const char * {
	switch (type) {
	case GL_DEBUG_TYPE_ERROR: return "Error";
		break;
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: return "Deprecated Behaviour";
		break;
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR: return "Undefined Behaviour";
		break;
	case GL_DEBUG_TYPE_PERFORMANCE: return "Performance";
		break;
	case GL_DEBUG_TYPE_PORTABILITY: return "Portability";
		break;
	case GL_DEBUG_TYPE_MARKER: return "Marker";
		break;
	case GL_DEBUG_TYPE_PUSH_GROUP: return "Push Group";
		break;
	case GL_DEBUG_TYPE_POP_GROUP: return "Pop Group";
		break;
	default:
	case GL_DEBUG_TYPE_OTHER: return "Other";
		break;
	}
	assert(false);
	return "unknown";
}

#pragma clang diagnostic push
#pragma ide diagnostic ignored "bugprone-easily-swappable-parameters"
void static fggl_ogl_to_spdlog(GLenum source, GLenum type, unsigned int msgId, GLenum severity, GLsizei  /*length*/,
							   const char *message, const void * /*userParam*/) {
	std::string fmt = "[GL] {}, {}: [{}]: {}";

	const auto *const sourceStr = fggl_ogl_source(source);
	const auto *const typeStr = fggl_ogl_type(type);

	// table 20.3, GL spec 4.5
	switch (severity) {
	case GL_DEBUG_SEVERITY_HIGH: spdlog::error(fmt, sourceStr, typeStr, msgId, message);
		break;
	default:
	case GL_DEBUG_SEVERITY_MEDIUM: spdlog::warn(fmt, sourceStr, typeStr, msgId, message);
		break;
	case GL_DEBUG_SEVERITY_LOW: spdlog::info(fmt, sourceStr, typeStr, msgId, message);
		break;
	case GL_DEBUG_SEVERITY_NOTIFICATION: spdlog::debug(fmt, sourceStr, typeStr, msgId, message);
		break;
	}
}
#pragma clang diagnostic pop

}

/**
 * Future optimisations:
 *   recommended approach is to group stuff in to as few vao as possible - this
 * will do one vao per mesh, aka bad. Add support for instanced rendering
 * (particles) Look at packing vertex data in better ways (with profiling)
 *   Support shader specialisation (ie, dynamic/streamed data)
 *   Follow best recommendations for Vertex attributes (ie normals not using
 * floats, bytes for colour)
 *
 * Future features:
 *   Add support for models with weights (for animations)
 *   OpenGL ES for the FOSS thinkpad users who can't run anything even remotely
 * modern
 */
namespace fggl::gfx {

	using data::Mesh2D;
	using data::Vertex2D;

	GlRenderToken setupVertex2D() {
		GlRenderToken token{};
		glGenVertexArrays(1, &token.vao);
		glBindVertexArray(token.vao);

		glGenBuffers(2, token.buffs);

		glBindBuffer(GL_ARRAY_BUFFER, token.buffs[0]);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex2D),
							  reinterpret_cast<void *>(offsetof(Vertex2D, position)));
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex2D),
							  reinterpret_cast<void *>(offsetof(Vertex2D, colour)));

		glBindVertexArray(0);
		return token;
	}

	OpenGL4Backend::OpenGL4Backend(const Window &owner) : fggl::gfx::Graphics() {
		// initialise GLEW, or fail
		GLenum err = glewInit();
		if (GLEW_OK != err) {
			throw std::runtime_error("couldn't init glew");
		}

		// OpenGL debug Support
		GLint flags = 0;
		glGetIntegerv(GL_CONTEXT_FLAGS, &flags);
		if ( (flags & GL_CONTEXT_FLAG_DEBUG_BIT) == GL_CONTEXT_FLAG_DEBUG_BIT) { // NOLINT(hicpp-signed-bitwise)
			spdlog::info("enabling OpenGL debug output");
			glEnable(GL_DEBUG_OUTPUT);
			glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
			glDebugMessageCallback(fggl_ogl_to_spdlog, nullptr);
			glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
		}

		// setup the viewport based on the window's framebuffer
		auto fbSize = owner.frameSize();
		glViewport(0, 0, fbSize.x, fbSize.y);

		// setup the shader cache
		auto &locator = util::ServiceLocator::instance();
		auto storage = locator.get<data::Storage>();
		m_cache = std::make_unique<ShaderCache>(storage);

		// setup 2D rendering system
		m_token2D = setupVertex2D();
		ShaderConfig shader2DConfig = ShaderFromName("shader2D");
		m_cache->load(shader2DConfig);

	};

	void OpenGL4Backend::clear() {
		glClearColor(0.0F, 0.0F, 0.0F, 0.0F);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	inline static void add_mesh_triangle(data::Mesh2D& mesh, const std::vector<data::Vertex2D>& verts) {
		assert( verts.size() == 3);
		for( const auto& vert : verts ) {
			auto idx = mesh.add_vertex(vert);
			mesh.add_index(idx);
		}
	}

	static void generateMesh(const gfx::Paint &paint, Mesh2D &mesh) {
		for (const auto &cmd : paint.cmds()) {
			auto path = cmd.path;

			std::vector<data::Vertex2D> verts;
			math::vec3 colour{1.0F, 1.0F, 1.0F};
			auto idx = 0;
			auto colourIdx = 0;

			for (auto &type : path.m_types) {
				if (type == PathType::PATH) {
					verts.push_back({.position = path.m_points[idx++], .colour = colour});
				} else if (type == PathType::MOVE) {
					// polygon finished
					if (verts.size() < 3) {
						// empty, point, or line
						// TODO deal with whatever I'm meant to do with this...
					} else if (verts.size() == 3) {
						// triangle
						add_mesh_triangle(mesh, verts);
					} else {
						// polygon
						math::fanTriangulation(verts, mesh);
					}

					verts.clear();
					verts.push_back({.position = path.m_points[idx++], .colour = colour});
				} else if (type == PathType::COLOUR) {
					colour = path.m_colours[colourIdx++];
				} else {
					// unsupported type
				}
			}

			if (!verts.empty()) {
				math::fanTriangulation(verts, mesh);
			}
		}
	}

	void OpenGL4Backend::draw2D(const gfx::Paint &paint) {
		// generate the mesh from a paint command list
		data::Mesh2D mesh;
		generateMesh(paint, mesh);

		// render the resulting mesh
		auto shader2D = m_cache->get("shader2D");

		glUseProgram(shader2D);

		auto projMat = glm::ortho(0.0f, 1920.0f, 0.0f, 1080.f);
		glUniformMatrix4fv(glGetUniformLocation(shader2D, "projection"), 1, GL_FALSE,
						   glm::value_ptr(projMat));

		glBindVertexArray(m_token2D.vao);

		glBindBuffer(GL_ARRAY_BUFFER, m_token2D.buffs[0]);
		glBufferData(GL_ARRAY_BUFFER, mesh.vertexList.size() * sizeof(Vertex2D),
					 mesh.vertexList.data(), GL_DYNAMIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_token2D.buffs[1]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER,
					 mesh.indexList.size() * sizeof(uint32_t), mesh.indexList.data(),
					 GL_DYNAMIC_DRAW);

		glDrawElements(GL_TRIANGLES, mesh.indexList.size(), GL_UNSIGNED_INT,
					   (GLvoid *) 0);

		glBindVertexArray(0);
		glUseProgram(0);
		// glDisable(GL_PRIMITIVE_RESTART);
	}

	void OpenGL4Backend::resize(int width, int height) {
		glViewport(0, 0, width, height);
	}

	template<typename T>
	static GLuint createArrayBuffer(std::vector<T> &vertexData) {
		GLuint buffId;
		glGenBuffers(1, &buffId);
		glBindBuffer(GL_ARRAY_BUFFER, buffId);
		glBufferData(GL_ARRAY_BUFFER, vertexData.size() * sizeof(T),
					 vertexData.data(), GL_STATIC_DRAW);
		return buffId;
	}

	static GLuint createIndexBuffer(std::vector<uint32_t> &indexData) {
		GLuint buffId;
		glGenBuffers(1, &buffId);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffId);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, indexData.size() * sizeof(uint32_t),
					 indexData.data(), GL_STATIC_DRAW);
		return buffId;
	}

	GlRenderToken MeshRenderer::upload(fggl::data::Mesh &mesh) {
		GlRenderToken token{};
		glGenVertexArrays(1, &token.vao);
		glBindVertexArray(token.vao);

		token.buffs[0] = createArrayBuffer<fggl::data::Vertex>(mesh.vertexList());
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(
			0, 3, GL_FLOAT, GL_FALSE, sizeof(fggl::data::Vertex),
			reinterpret_cast<void *>(offsetof(fggl::data::Vertex, posititon)));
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(
			1, 3, GL_FLOAT, GL_FALSE, sizeof(fggl::data::Vertex),
			reinterpret_cast<void *>(offsetof(fggl::data::Vertex, normal)));

		token.buffs[1] = createIndexBuffer(mesh.indexList());
		token.idxOffset = 0;
		token.idxSize = mesh.indexCount();
		token.restartVertex = mesh.restartVertex;
		glBindVertexArray(0);

		return token;
	}

// TODO(webpigeon): this shouldn't be hard-coded
	constexpr glm::vec3 DEFAULT_LIGHTPOS = glm::vec3(20.0F, 20.0F, 15.0F);

	void MeshRenderer::render(fggl::ecs3::World &ecs, ecs3::entity_t camera,
							  float dt) {
		if (camera == ecs::NULL_ENTITY) {
			spdlog::warn("tried to render a scene, but no camera exists!");
			return;
		}

		auto entities = ecs.findMatching<GlRenderToken>();
		if (entities.empty()) {
			spdlog::warn("asked to render, but no entities are renderable");
			return;
		}

		total += dt;

		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		glEnable(GL_DEPTH_TEST);

		// camera logic
		auto *const camTransform = ecs.get<math::Transform>(camera);
		auto *const camComp = ecs.get<gfx::Camera>(camera);
		glm::mat4 proj = glm::perspective(camComp->fov, camComp->aspectRatio,
										  camComp->nearPlane, camComp->farPlane);
		glm::mat4 view =
			glm::lookAt(camTransform->origin(), camComp->target, camTransform->up());

		// lighting
		glm::vec3 lightPos = DEFAULT_LIGHTPOS;

		// TODO(webpigeon): better performance if grouped by vao first
		// TODO(webpigeon): the nvidia performance presentation said I shouldn't use
		// uniforms for large data
		for (auto &entity : entities) {
			const auto &transform = ecs.get<fggl::math::Transform>(entity);
			const auto &mesh = ecs.get<GlRenderToken>(entity);

			glm::mat4 model = transform->model();
			//	    model = glm::rotate(model, glm::radians(total/2048.0f * 360.0f),
			//glm::vec3(0.0f,1.0f,0.0f));

			auto shader = mesh->pipeline;
			glUseProgram(shader);

			glUniformMatrix4fv(glGetUniformLocation(shader, "model"), 1, GL_FALSE,
							   glm::value_ptr(model));
			glUniformMatrix4fv(glGetUniformLocation(shader, "view"), 1, GL_FALSE,
							   glm::value_ptr(view));
			glUniformMatrix4fv(glGetUniformLocation(shader, "projection"), 1, GL_FALSE,
							   glm::value_ptr(proj));

			// lighting
			GLint lightID = glGetUniformLocation(shader, "lightPos");
			if (lightID != -1) {
				glUniform3fv(lightID, 1, glm::value_ptr(lightPos));
			}

			glBindVertexArray(mesh->vao);

			if (mesh->renderType == GlRenderType::triangle_strip) {
				glEnable(GL_PRIMITIVE_RESTART);
				glPrimitiveRestartIndex(mesh->restartVertex);
			}

			glDrawElements(mesh->renderType, mesh->idxSize, GL_UNSIGNED_INT,
						   reinterpret_cast<void *>(mesh->idxOffset));
			if (mesh->renderType == GlRenderType::triangle_strip) {
				glDisable(GL_PRIMITIVE_RESTART);
			}
		}

		glBindVertexArray(0);
	}

}  // namespace fggl::gfx
